/*
 * $Id: //trid/release/pam_otp_auth-3.2.2/otp_pw_valid.c#1 $
 *
 * Passcode verification function (otpd client) for pam_otp_auth.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 *
 *  For alternative licensing, contact licensing@tri-dsystems.com.
 *
 * Copyright 2006 TRI-D Systems, Inc.
 */

#include "ident.h"
RCSID("$Id: //trid/release/pam_otp_auth-3.2.2/otp_pw_valid.c#1 $")

#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <syslog.h>
#include <unistd.h>
#include <security/pam_appl.h>	/* PAM_* */

#include "extern.h"
#include "otp.h"
#include "otp_pw_valid.h"

/* transform otp_pw_valid() return code into a pam return code */
static int
otprc2pamrc(int rc)
{
  switch (rc) {
    case OTP_RC_OK:			return PAM_SUCCESS;
    case OTP_RC_USER_UNKNOWN:		return PAM_USER_UNKNOWN;
    case OTP_RC_AUTHINFO_UNAVAIL:	return PAM_AUTHINFO_UNAVAIL;
    case OTP_RC_AUTH_ERR:		return PAM_AUTH_ERR;
    case OTP_RC_MAXTRIES:		return PAM_MAXTRIES;
    case OTP_RC_SERVICE_ERR:		return PAM_SERVICE_ERR;
    default:				return PAM_SERVICE_ERR;
  }
}

/*
 * Handle SIGPIPE so we don't exit on write() to otpd.
 *
 * It's not possible to guarantee correctness in a multithreaded app.
 * We have to handle, not just block, SIGPIPE, because many OS's send
 * SIGPIPE to any thread (considering it as a process-directed signal)
 * instead of the thread which caused it.  This is due to bad language
 * in POSIX.  See
 * <http://groups.google.com/groups?threadm=3f2fb931@usenet01.boi.hp.com>.
 * If we just blocked SIGPIPE, another thread might get it.
 *
 * Even if we handle SIGPIPE, there's a TOCTOU race.  It seems extremely
 * unlikely that an app would be written such that this will break.  It
 * would have to call pam_authenticate() in one thread, then while the
 * authentication is happening, start another thread and change SIGPIPE
 * handling in that thread.
 *
 * We also have the risk that a multi-threaded app may run multiple auths
 * concurrently.  We can't put a mutex around restore_sigpipe_sa, because
 * that will require bringing in libpthread (in most cases) and this will
 * break non-threaded apps (in most cases).  There's no way to handle this.
 */
static struct sigaction sigpipe_sa;
static int restore_sigpipe_sa;
static void
sigpipe_wrap(void)
{
  struct sigaction sa;

  /* already been here? */
  if (restore_sigpipe_sa)
    return;

  (void) sigaction(SIGPIPE, NULL, &sigpipe_sa);
  /* do nothing if already handled */
  if (sigpipe_sa.sa_handler != SIG_DFL)
    return;

  /* ignore SIGPIPE */
  sa.sa_handler = SIG_IGN;
  sa.sa_flags = 0;
  (void) sigemptyset(&sa.sa_mask);	/* not really needed */
  (void) sigaction(SIGPIPE, &sa, NULL);
  restore_sigpipe_sa = 1;
}

static void
sigpipe_unwrap(void)
{
  if (restore_sigpipe_sa)
    (void) sigaction(SIGPIPE, &sigpipe_sa, NULL);
  restore_sigpipe_sa = 0;
}

/*
 * Test for passcode validity by asking otpd.
 *
 * If challenge is supplied, it is used to generate the card response
 * against which the passcode will be compared.  If challenge is not
 * supplied, or if the comparison fails, synchronous responses are
 * generated and tested.  NOTE: for async authentications, sync mode
 * responses are still considered valid!  (Assuming module configuration
 * allows sync mode.)
 *
 * Returns one of the PAM_* codes.
 */
int
otp_pw_valid(const char *username, const char *passcode,
             const char *challenge, const otp_option_t *opt)
{
  otp_request_t	otp_request;
  otp_reply_t	otp_reply;
  int rc;

  if (strlen(username) > OTP_MAX_USERNAME_LEN) {
    syslog(LOG_NOTICE, "pam_otp_auth: username [%s] too long", username);
    return PAM_AUTH_ERR;
  }
  /* we already know challenge is short enough */

  otp_request.version = 2;
  (void) strcpy(otp_request.username, username);
  (void) strcpy(otp_request.challenge, challenge);
  otp_request.pwe.pwe = PWE_PAP;

  if (strlen(passcode) > OTP_MAX_PASSCODE_LEN) {
    syslog(LOG_NOTICE, "pam_otp_auth: passcode for [%s] too long", username);
    return PAM_AUTH_ERR;
  }
  (void) strcpy(otp_request.pwe.u.pap.passcode, passcode);

  /* last byte must also be a terminator so otpd can verify length easily */
  otp_request.username[OTP_MAX_USERNAME_LEN] = '\0';
  otp_request.challenge[OTP_MAX_CHALLENGE_LEN] = '\0';
  otp_request.pwe.u.pap.passcode[OTP_MAX_PASSCODE_LEN] = '\0';

  otp_request.allow_sync = opt->allow_sync;
  otp_request.allow_async = opt->allow_async;
  otp_request.challenge_delay = 0;
  otp_request.resync = 1;

  sigpipe_wrap();
  rc = otprc2pamrc(otp_verify(opt, &otp_request, &otp_reply));
  sigpipe_unwrap();

  return rc;
}

/*
 * Verify an otp by asking otpd.
 * Returns an OTP_* code, or -1 on system failure.
 * Fills in reply.
 */
static int
otp_verify(const otp_option_t *opt,
           const otp_request_t *request, otp_reply_t *reply)
{
  int fd, rc;

  /*
   * We expect single round trips so we open and close fd
   * on each call.  If a pam app uses other than the fork/exec
   * model, we might want to change this.
   */
  if ((fd = otp_connect(opt->otpd_rp)) == -1)
    return -1;

  if ((rc = otp_write(fd, (const char *) request, sizeof(*request))) != 0)
      return -1;

  if ((rc = otp_read(fd, (char *) reply, sizeof(*reply))) != sizeof(*reply))
      return -1;

  (void) close(fd);

  /* validate the reply */
  if (reply->version != 1) {
    syslog(LOG_NOTICE, "pam_otp_auth: otpd reply for [%s] invalid "
                       "(version %d != 1)",
           request->username, reply->version);
    return -1;
  }

  if (reply->passcode[OTP_MAX_PASSCODE_LEN] != '\0') {
    syslog(LOG_NOTICE, "pam_otp_auth: otpd reply for [%s] invalid (passcode)",
           request->username);
    return -1;
  }

  return reply->rc;
}

/* connect to otpd and return fd */
static int
otp_connect(const char *path)
{
  int fd;
  struct sockaddr_un sa;
  size_t sp_len;		/* sun_path length (strlen) */

  /* setup for unix domain socket */
  sp_len = strlen(path);
  if (sp_len > sizeof(sa.sun_path) - 1) {
    syslog(LOG_ERR, "pam_otp_auth: %s: rendezvous point name too long",
           __func__);
    return -1;
  }
  sa.sun_family = AF_UNIX;
  (void) strcpy(sa.sun_path, path);
    
  /* connect to otpd */
  if ((fd = socket(PF_UNIX, SOCK_STREAM, 0)) == -1) {
    syslog(LOG_ERR, "pam_otp_auth: %s: socket: %s", __func__, strerror(errno));
    return -1;
  }
  if (connect(fd, (struct sockaddr *) &sa,
              sizeof(sa.sun_family) + sp_len) == -1) {
    syslog(LOG_ERR, "pam_otp_auth: %s: connect(%s): %s",
           __func__, path, strerror(errno));
    (void) close(fd);
    return -1;
  }
  return fd;
}
