/*
 * $Id: //trid/release/pam_otp_auth-3.2.2/otp_support.c#1 $
 *
 * Support functions for pam_otp_auth.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 *
 *  For alternative licensing, contact licensing@tri-dsystems.com.
 *
 * Copyright 2001,2002  Google, Inc.
 * Copyright 2005,2006 TRI-D Systems, Inc.
 */

#include "ident.h"
RCSID("$Id: //trid/release/pam_otp_auth-3.2.2/otp_support.c#1 $")

#include <errno.h>
#include <stdlib.h>
#include <security/pam_appl.h>
#include <security/pam_modules.h>
#include <string.h>
#include <sys/stat.h>
#include <syslog.h>
#include <unistd.h>

#include "extern.h"
#include "otp.h"	/* OTP_MAX_CHALLENGE_LEN */

#define DLOGOPT \
if (opt->debug) \
  syslog(LOG_DEBUG, "pam_otp_auth: parse_opts: set %s", argv[i])

/* The "" 's below are to ensure that b is a literal string, not a var */
#define strprefix(a, b) ( !strncmp(a, b, sizeof("" b "") - 1) )
#define strsize(b) (sizeof("" b "") - 1)

/*
 * Parse options and return them in opt.
 * Returns PAM_SUCCESS on success, PAM_SERVICE_ERR otherwise.
 * Caller must call free_opts() after calling parse_opts() (regardless of
 * return value).
 */
int
otp_parse_opts(int argc, const char **argv, const char *service,
               otp_option_t *opt)
{
  int i;
  char *p;

  /* Set defaults */
  (void) memset(opt, 0, sizeof(*opt));
  opt->otpd_rp		= OTP_OTPD_RP;
  opt->fast_prompt	= strdup(OTP_FAST_PROMPT);
  opt->chal_prompt	= strdup(OTP_CHALLENGE_PROMPT);
  opt->chal_len		= 6;
  opt->fast_sync	= 1;
  opt->allow_sync	= 1;
  opt->allow_async	= 0;
  opt->chal_req		= OTP_CHALLENGE_REQ;
  opt->useauthtok	= 0;
  opt->debug		= 0;

  if (!opt->chal_prompt) {
    syslog(LOG_CRIT, "pam_otp_auth(%s): %s: out of memory!", service, __func__);
    return PAM_SERVICE_ERR;
  }
  if (!opt->fast_prompt) {
    syslog(LOG_CRIT, "pam_otp_auth(%s): %s: out of memory!", service, __func__);
    return PAM_SERVICE_ERR;
  }

  for (i = 0; i < argc; ++i) {
    if (!strcmp(argv[i], "debug")) {
      opt->debug = 1;
      DLOGOPT;

    } else if (!strncmp(argv[i], "otpd_rp=/", 9)) {
      opt->otpd_rp = (char *) &argv[i][8];
      DLOGOPT;

    } else if (!strncmp(argv[i], "fast_prompt=", 12)) {
      opt->fast_prompt = strdup(&argv[i][12]);
      if (!opt->fast_prompt) {
        syslog(LOG_CRIT, "pam_otp_auth(%s): %s: out of memory!",
               service, __func__);
        return PAM_SERVICE_ERR;
      }
      /* "_" -> ' ' and backslash escapes */
      while ((p = strchr(opt->fast_prompt, '_')))
        *p = ' ';
      while ((p = strchr(opt->fast_prompt, '\\'))) {
        if (*++p == 'n')
          *--p = '\n';
        else
          p -= 2;
        (void) memmove(p + 1, p + 2, strlen(p + 1));
      }
      DLOGOPT;

    } else if (!strncmp(argv[i], "challenge_prompt=", 17)) {
      /* Enforce a single "%" sequence, which must be "%s". */
      p = strchr(&argv[i][17], '%');
      if (p == NULL || p != strrchr(&argv[i][17], '%') ||
          strncmp(p,"%s", 2)) {
        syslog(LOG_ERR, "pam_otp_auth: invalid challenge text, using %s",
               opt->chal_prompt);
      } else {
        opt->chal_prompt = strdup(&argv[i][17]);
        if (!opt->chal_prompt) {
          syslog(LOG_CRIT, "pam_otp_auth(%s): %s: out of memory!",
                 service, __func__);
          return PAM_SERVICE_ERR;
        }
        /* "_" -> ' ' and backslash escapes */
        while ((p = strchr(opt->chal_prompt, '_')))
          *p = ' ';
        while ((p = strchr(opt->chal_prompt, '\\'))) {
          if (*++p == 'n')
            *--p = '\n';
          else
            p -= 2;
          (void) memmove(p + 1, p + 2, strlen(p + 1));
        }
      }
      DLOGOPT;

    } else if (!strncmp(argv[i], "challenge_length=", 17)) {
      opt->chal_len = (int) strtoul(&argv[i][17], NULL, 10);
      if (opt->chal_len < 5 || opt->chal_len > OTP_MAX_CHALLENGE_LEN) {
        opt->chal_len = 6;
        syslog(LOG_ERR, "pam_otp_auth: invalid challenge length, range 5-%d, "
                        "using default of 6", OTP_MAX_CHALLENGE_LEN);
      }
      DLOGOPT;

    } else if (!strncmp(argv[i], "fast_sync=", 10)) {
      opt->fast_sync = (int) strtoul(&argv[i][10], NULL, 10);
      DLOGOPT;

    } else if (!strncmp(argv[i], "allow_sync=", 11)) {
      opt->allow_sync = (int) strtoul(&argv[i][11], NULL, 10);
      DLOGOPT;

    } else if (!strncmp(argv[i], "allow_async=", 12)) {
      opt->allow_async = (int) strtoul(&argv[i][12], NULL, 10);
      DLOGOPT;

    } else if (!strncmp(argv[i], "challenge_req=", 14)) {
      opt->chal_req = (char *) &argv[i][14];
      DLOGOPT;

    } else if (!strncmp(argv[i], "useauthtok", 10)) {
      opt->useauthtok = 1;
      DLOGOPT;

    } else {
      syslog(LOG_ERR, "pam_otp_auth: unknown option %s", argv[i]);
      return PAM_SERVICE_ERR;
    }
  }

  if (!(opt->allow_async || opt->allow_sync)) {
    syslog(LOG_ERR, "pam_otp_auth: at least one of {allow_async, allow_sync} "
                    "must be set");
    return PAM_SERVICE_ERR;
  }
  if (opt->fast_sync && !opt->allow_sync) {
    opt->fast_sync = 0;
    syslog(LOG_ERR, "pam_otp_auth: fast_sync is yes, "
                    "but allow_sync is no; disabling fast_sync");
  }

  return PAM_SUCCESS;
}


/*
 * Caller must call free_opts() after calling parse_opts() (regardless of
 * return value).
 */
void
otp_free_opts(otp_option_t *opt)
{
  if (opt->chal_prompt) {
    free(opt->chal_prompt);
    opt->chal_prompt = NULL;
  }
  if (opt->fast_prompt) {
    free(opt->fast_prompt);
    opt->fast_prompt = NULL;
  }
}


/*
 * Get info from the user. Disallow null responses (regardless of flags).
 * response gets allocated and filled in on successful return. Caller
 * is responsible for freeing it.
 */
int
otp_interact(pam_handle_t *pamh, char *prompt, int type, char **response)
{
  int pamret;
  struct pam_message	msg, *pmsg;
  struct pam_response	*resp = NULL;
  struct pam_conv	*conv;
  void			*vconv;	/* void * due to gcc aliasing deficiency */

  if ((pamret = pam_get_item(pamh, PAM_CONV, (PAM_GET_CONST void **) &vconv)))
    return pamret;
  conv = (struct pam_conv *) vconv;

  /* set up conversation call */
  pmsg = &msg;
  msg.msg_style = type;
  msg.msg = prompt;

  if ((pamret = conv->conv(1, (PAM_CONV_CONST struct pam_message **) &pmsg,
                           &resp, conv->appdata_ptr)))
    return pamret;

  /* Caller should ignore errors for non-response conversations */
  if (!resp)
    return PAM_CONV_ERR;

  if (!(resp->resp && resp->resp[0])) {
    free(resp);
    return PAM_AUTH_ERR;
  }

  *response = resp->resp;
  free(resp);
  return pamret;
}
